%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% learn_structure_mcdn_simple:
%
% This does not iterate: it first fits a model with no bi-directed edges,
% then add edges for each pair where the marginal likelihood says so.
%
% Input:
%
% - Y: the data, binary encoded as -1s and 1s
% - Y_counts: bivariate sufficient statistics
% - X_children: list of 2 containing the of latent children
% - p_G: probability of anyedge existing
% - K: maximum number of optimization iterations
% - sd_r_prior: prior standard deviation for edge-copula parameters
% - X_v: grid values shared by all latent variables
% - X_ws: cell structure containing latent weights for a range of possible 
%         correlations
% - rho_domain: possible correlation values
% - M: number of importance samples
%
% Output:
% 
% - H: bi-directed structure
% - L: loading matrix
% - rho_X: latent correlation (might not be positive definite)
% - final_score: pairwise composite likelihood score
% - pair_scores: detailed scores per pair
% - original_tabu_edges: indicates which coefficients were fixed by
%                        initialization procedure
% - tabu_values: the corresponding values used in the initialization
% - num_iter: final number of iterations
%
% Created by: Ricardo Silva, London, 06/03/2012
% University College London
% Current version: 16/06/2012

function [H L rho_X final_score pair_scores] = ...
   learn_structure_mcdn_simple(Y, Y_counts, X_children, p_G, ...
                                sd_r_prior, X_v, X_ws, rho_domain, M)

% Initialization %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[N num_y] = size(Y);
num_x = length(X_children);
H = zeros(num_y); 
Y_parents = zeros(num_y, 1);
for x = 1:num_x
  for y = X_children{x} 
    Y_parents(y) = x;
  end
end

prob_zero = mean(Y == -1)';

L = robust_initialize_pure_model(Y, X_children, X_v, X_ws, ...
                                 rho_domain, zeros(num_y, 1), zeros(num_y, 2), prob_zero);
        
num_regions = num_x * (num_x - 1) / 2;
regions = zeros(2, num_regions); x_count = 1;
for x1 = 1:num_x    
  for x2 = (x1 + 1):num_x
    regions(:, x_count) = [x1; x2]; x_count = x_count + 1;
  end
end

[cop_theta_prior prior_w] = build_distr_approx(M, [0 sd_r_prior^2 Inf]);

options_optim = optimset('GradObj', 'off', 'LargeScale', 'off', ...
                         'Display', 'iter', 'DerivativeCheck', 'off', ...
                         'TolFun', 1.e-3);

C_region = cell(num_regions, 1);
is_w_region = cell(num_regions, 1);
for i = 1:num_regions
  C_region{i} = zeros(num_y, num_y, M);
  is_w_region{i} = zeros(num_y, num_y, M);
  for j = 1:num_y
    for k = 1:num_y
      C_region{i}(j, k, :) = cop_theta_prior;
      is_w_region{i}(j, k, :) = prior_w;
    end
  end
end    

scale = 25;
fix_rho = true;

num_inner_i = 2; 

% Single shot %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


  fprintf('> Optimizing: single shot\n')
  
  expo_y_all = get_expo_y(1:num_y, H);
  
  % Parameter point estimate

  rho_X = fit_rho_X_fixed_exp(L, Y_counts, ...
            X_children, H, regions, expo_y_all, C_region, is_w_region, ...
            X_v, X_ws, rho_domain, scale, zeros(num_x), zeros(num_x));
  for inner_i = 1:num_inner_i
    theta_0 = [L(:, 1); L(:, 2)]';
    theta = fminunc(@(theta)score_parameters_cllik_mcdn_fixed_exp(theta, ...
             rho_X, rho_domain, Y_counts, Y_parents, X_children, H, ...
              regions, expo_y_all, C_region, is_w_region, X_v, X_ws, true, ...
              scale, zeros(num_y, 1), zeros(num_y, 2), -N), ...
                 theta_0, options_optim);
    L = reshape(theta, num_y, 2);
    rho_X = fit_rho_X_fixed_exp(L, Y_counts, ...
              X_children, H, regions, expo_y_all, C_region, is_w_region, ...
              X_v, X_ws, rho_domain, scale, zeros(num_x), zeros(num_x));
  end
  
  % Add edges
  
  final_H = zeros(num_y);
  for y1 = 1:num_y
    for y2 = (y1 + 1):num_y
      H(y1, y2) = 1; H(y2, y1) = 1;
      expo_y_all = get_expo_y(1:num_y, H);
      score1 = score_edge_cllik_mcdn_priorbased(y1, y2, L, ...
                  rho_X, rho_domain, Y_counts, H, Y_parents, expo_y_all, ...
                  cop_theta_prior, prior_w, X_v, X_ws, fix_rho, scale, p_G);
      H(y1, y2) = 0; H(y2, y1) = 0;
      expo_y_all = get_expo_y(1:num_y, H);
      score0 = score_edge_cllik_mcdn_priorbased(y1, y2, L, ...
                  rho_X, rho_domain, Y_counts, H, Y_parents, expo_y_all, ...
                  cop_theta_prior, prior_w, X_v, X_ws, fix_rho, scale, p_G); 
      if score1 > score0
        final_H(y1, y2) = 1; final_H(y2, y1) = 1;
        fprintf('* Adding edge %d <--> %d\n', y1, y2);
      end
    end
  end
  H = final_H;

  % Wrap-up
  
  fprintf('FINISHED\n');
  final_score = 0; expo_y_all = get_expo_y(1:num_y, H);
  pair_scores = zeros(num_y);
  for y1 = 1:num_y
    for y2 = (y1 + 1):num_y  
      pair_scores(y1, y2) = ...
          score_edge_cllik_mcdn_priorbased(y1, y2, L, ...
                  rho_X, rho_domain, Y_counts, H, Y_parents, expo_y_all, ...
                  cop_theta_prior, prior_w, X_v, X_ws, true, scale, p_G, false); 
      pair_scores(y2, y1) = pair_scores(y1, y2);
      final_score = final_score + pair_scores(y1, y2);            
    end
  end
  final_score = final_score / N;
  fprintf('Final score = %f\n', final_score);

   function y = intercept(theta, probs)
      y = - sqrt(theta.^2 + 1) .* norminv(probs);
   end

end
