%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% score_edge_cllik_mcdn_priorbased:
%
% Score the marginal likelihood for a composite likelihood score with and 
% without an edge. Exponents are fixed instead of random. Marginalization 
% is done over the prior.
%
% Input:
%
% - y_target_1, y_target_2: variables in which the edge is being evaluated
% - L: loadings
% - rho_X: latent correlations
% - rho_domain: space of possible correlation values for latent variables
% - Y_counts: pairwise sufficient statistics
% - X_children: same information as Y_parents, but indexed by the latents
% - G: bi-directed structure
% - Y_exp: samples for the measurement exponents for each region
% - X_v: latent points
% - X_ws: corresponding weights, organized as a matrix
%   
% Output:
% 
% - score1: score with the edge in
% - score0: score with the edge out
%
% Created by: Ricardo Silva, London, 09/03/2012
% University College London
% Current version: 09/03/2012

function score = ...
   score_edge_cllik_mcdn_priorbased(y_target_1, y_target_2, L, rho_X, ...
      rho_domain, Y_counts, G, Y_parents, expo_y, cop_theta, cop_theta_w, ...
      X_v, X_ws, rho_fixed, scale, p_G, include_neighbors)

if nargin < 17
  include_neighbors = true;
end

M = length(cop_theta);        
num_quad = size(X_v, 1);

if G(y_target_1, y_target_2) == 1
  score = log(p_G);
else
  score = log(1 - p_G);  
end

% Calculate function

bin_values = [0 0; 0 1; 1 0; 1 1]'; signs = [1 -1 -1 1];
M_ones  = ones(M, num_quad);
M1_ones = ones(1, num_quad);
p1 = 1;
  
if include_neighbors
  sp_y_1 = find(G(y_target_1, :) == 1);
  sp_y_2 = find(G(y_target_2, :) == 1);
  range_y_1 = [y_target_1 y_target_2];
  range_y_2_1 = union(y_target_2, sp_y_1);
  range_y_2_2 = setdiff(sp_y_2, y_target_1);
else
  range_y_1 = y_target_1;
  range_y_2_1 = y_target_2;
end

for y_i = range_y_1
            
  if y_i == y_target_1
    range_y_2 = range_y_2_1;
  else
    range_y_2 = range_y_2_2;
  end
    
  for y_j = range_y_2
        
    if Y_parents(y_j) == Y_parents(y_i)
      p2 = 1; 
      if rho_fixed
        X_w_choice = X_ws(:, round(0.5 * length(rho_domain)));
      else
        X_w_choice = X_ws;
      end
    else
      p2 = 2;
      if rho_fixed
        rho_idx = find(rho_domain == rho_X(Y_parents(y_i), Y_parents(y_j)), 1);
        X_w_choice = X_ws(:, rho_idx);      
      else
        X_w_choice = X_ws;
      end          
    end
      
    U = cdf_probit(L([y_i y_j], :), [p1 p2], X_v);
                     
    if G(y_i, y_j) == 1
      rep_u_i = repmat(U(:, 1)', M, 1);  
      rep_u_j = repmat(U(:, 2)', M, 1);
      exp_i = expo_y(y_i, y_j);
      exp_j = expo_y(y_j, y_i);
      rep_ct = repmat(cop_transform(cop_theta, scale), 1, num_quad);
      M_iter = M;
      cop_theta_w_choice = cop_theta_w;
    else
      M_iter = 1;
      cop_theta_w_choice = 1;
    end
    pmf_rho = zeros(M_iter, 1);
    
    % SCORE LIKELIHOOD %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      
    for v = bin_values % Iterate over all instantiations of {Y_i, Y_j}
                                
      D = zeros(M_iter, num_quad);
      z_pos = 1;
        
      for z = bin_values % CDF to PMF transformation
            
        i_zv = v(1) - z(1); j_zv = v(2) - z(2); 
        if i_zv < 0 || j_zv < 0, continue; end                                   

        % SCORE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        if G(y_i, y_j) == 1
          if i_zv == 0                
            u_i   = rep_u_i.^exp_i;
            u_i0  = rep_u_i.^(1 - exp_i);  
          else
            u_i   = M_ones;
            u_i0  = M_ones;
          end
          if j_zv == 0                
            u_j   = rep_u_j.^exp_j;
            u_j0  = rep_u_j.^(1 - exp_j);
          else
            u_j   = M_ones;
            u_j0  = M_ones;
          end
          aux = 1 + (exp(-rep_ct .* u_i) - 1) .* (exp(-rep_ct .* u_j) - 1) ./ ...
                    (exp(-rep_ct) - 1); 
          C_aux = -log(aux) ./ rep_ct;
          D = D + signs(z_pos) * (u_i0 .* u_j0 .* C_aux);
        else % G(y_i, y_j) == 0
          if i_zv == 0
            u_i0 = U(:, 1)';
          else
            u_i0 = M1_ones;
          end  
          if j_zv == 0
            u_j0 = U(:, 2)';
          else
            u_j0 = M1_ones;
          end
          D = D + signs(z_pos) * (u_i0 .* u_j0);
        end
          
        % ITERATE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
          
        z_pos = z_pos + 1;
          
      end % end of PMF transformation          

      N_ij = Y_counts{v(1) + 1, v(2) + 1}(y_i, y_j);
      pmf = mean(D * X_w_choice, 2);
      pmf_rho = pmf_rho + N_ij * log(pmf);
        
    end % end of {Y_i, Y_j} instantiations
      
    % SCORE PRIOR %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
    Y_c = cell(4, 1); for uu1 = 1:2, for uu2 = 1:2, Y_c{uu1, uu2} = Y_counts{uu1, uu2}(y_i, y_j); end; end
    m_pmf = max(pmf_rho);
    score = score + log(exp(pmf_rho - m_pmf)' * cop_theta_w_choice) + m_pmf;
    % Visualize it: plot(cop_transform(cop_theta, scale), exp(pmf_rho - max(pmf_rho)))
    
  end % end of Y_j calculations
    
end % end of Y_i calculations

