%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% score_parameters_cllik_mcdn_fixed_exp:
%
% Score loading and latent correlation parameters using composite 
% likelihood and samples for measurement dependence. This is done by 
% pairwise composite likelihood. Exponents within each factor are fixed.
% Regions and parent regions needed to be specified.
%
% Input:
%
% - theta: parameters. This corresponds to loading parameters 
% - rho_X: fixed latent correlations
% - Y_counts: pairwise sufficient statistics
% - X_children: same information as Y_parents, but indexed by the latents
% - G: bi-directed structure
% - regions: a list of pairs of latent variables defining the regions
% - Y_exp: samples for the measurement exponents for each region
% - C: samples for the copula parameters for each region
% - is_w: sample weights
% - X_v: latent points
% - X_ws: corresponding weights, organized as a matrix
% - rho_domain: space of possible correlation values for latent variables
% - fix_rho: if true, uses the rho values in "rho_X"
% - N: a constant to divide the function value by
%   
% Output:
% 
% - score: the corresponding  composite log-likelihood scaled by N
% - gr: the corresponding gradient scaled by N
%
% Created by: Ricardo Silva, London, 05/03/2012
% University College London
% Current version: 14/03/2012

function [score gr] = ...
   score_parameters_cllik_mcdn_fixed_exp(theta, rho_X, rho_domain, Y_counts, ...
       Y_parents, X_children, G, regions, expo_y, C, is_w, X_v, X_ws, fix_rho, ...
       scale, tabu_edges, tabu_values, N)

num_y = size(G, 1);
num_x = size(rho_X, 1);
num_quad = size(X_v, 1);

score = 0; gr = zeros(length(theta), 1);

L = zeros(num_y, 2); 
L(tabu_edges == 0, :) = reshape(theta, sum(tabu_edges == 0), 2); 
L(tabu_edges == 1, :) = tabu_values(tabu_edges == 1, :);

%L = reshape(theta, 2, num_y)'; % Unpack parameters

D_g = cell(4, 1); % Each cell is the M-by-num_quad derivative of each
                  % P(y_i, y_j | X, theta_ij) with respect to the local
                  % loadings. The joint combination of values for y_i and
                  % y_j corresponds to the possible 4 cells.

% Calculate function

bin_values = [0 0; 0 1; 1 0; 1 1]'; signs = [1 -1 -1 1];
M1_ones = ones(1, num_quad);
rep_X = cell(2, 1);
if ~fix_rho, X_w_choice = X_ws; end

n_r = 1;

region_weight = [1 / (num_x - 1) 1; 1 1 / (num_x - 1)];

for r = regions;

  M = size(C{n_r}, 3); % Number of samples
  rep_X{1} = repmat(X_v(:, 1)', M, 1); rep_X{2} = repmat(X_v(:, 2)', M, 1);
  M_ones = ones(M, num_quad);
  
  Y_scope = [X_children{r(1)}; X_children{r(2)}];
  local_Y_parents = [ones(length(X_children{r(1)}), 1); ...
                     2 * ones(length(X_children{r(2)}), 1)];
  U = cdf_probit(L(Y_scope, :), local_Y_parents, X_v);
  U_g = pdf_probit(L(Y_scope, :), local_Y_parents, X_v);
  num_y_r = length(Y_scope);  
    
  for y_i_idx = 1:num_y_r
    
    y_i = Y_scope(y_i_idx);    
    U_i = U(:, y_i_idx);
    U_g_i = U_g(:, y_i_idx);
    if y_i_idx <= length(X_children{r(1)})
      p1 = 1;
    else
      p1 = 2;
    end
    
    for y_j_idx = (y_i_idx + 1):num_y_r
        
      y_j = Y_scope(y_j_idx);      
      U_j = U(:, y_j_idx);
      U_g_j = U_g(:, y_j_idx);
      if G(y_i, y_j) == 1          
        exp_i = expo_y(y_i, y_j);
        exp_j = expo_y(y_j, y_i);
        %rep_u_i = repmat(U_i', M, 1);
          rep_u_i_exp_i = repmat(U_i'.^exp_i, M, 1);
          rep_u_i_exp_i_minus = repmat(U_i'.^(1 - exp_i), M, 1);
        rep_u_i_g = repmat(U_g_i', M, 1);
        %rep_u_j = repmat(U_j', M, 1);
          rep_u_j_exp_j = repmat(U_j'.^exp_j, M, 1);
          rep_u_j_exp_j_minus = repmat(U_j'.^(1 - exp_j), M, 1);
        rep_u_j_g = repmat(U_g_j', M, 1);
        rho = cop_transform(squeeze(C{n_r}(y_i_idx, y_j_idx, :)), scale);
        rep_rho = repmat(rho, 1, num_quad);
        M_iter = M;
        w_iter = squeeze(is_w{n_r}(y_i_idx, y_j_idx, :));
      else
        %rep_u_i = U_i';
        %rep_u_i_g = U_g_i';
        %rep_u_j = U_j';
        %rep_u_j_g = U_g_j';
        M_iter = 1;
        w_iter = 1;
      end
            
      if Y_parents(y_i) ~= Y_parents(y_j)
        if fix_rho
          rho_idx = find(rho_domain == rho_X(Y_parents(y_i), Y_parents(y_j)), 1);
          X_w_choice = X_ws(:, rho_idx);      
        end
        if p1 == 1, p2 = 2; else p2 = 1; end
      else
        if fix_rho
          X_w_choice = X_ws(:, round(0.5 * length(rho_domain)));
        end
        p2 = p1;
      end

      for v = bin_values % Iterate over all instantiations of {Y_i, Y_j}
                                
        D = zeros(M_iter, num_quad); 
        for g = 1:4, D_g{g} = zeros(M_iter, num_quad); end
        z_pos = 1;
        
        for z = bin_values % CDF to PMF transformation
            
          i_zv = v(1) - z(1); j_zv = v(2) - z(2); 
          if i_zv < 0 || j_zv < 0, continue; end                                   

          % SCORE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%          

          if G(y_i, y_j) == 1
            if i_zv == 0                
              %u_i   = rep_u_i.^exp_i;
              %u_i0  = rep_u_i.^(1 - exp_i);  
              u_i    = rep_u_i_exp_i;
              u_i0   = rep_u_i_exp_i_minus;
            else
              u_i   = M_ones;
              u_i0  = M_ones;
            end
            if j_zv == 0                
              %u_j   = rep_u_j.^exp_j;
              %u_j0  = rep_u_j.^(1 - exp_j);
              u_j    = rep_u_j_exp_j;
              u_j0   = rep_u_j_exp_j_minus;
            else
              u_j   = M_ones;
              u_j0  = M_ones;
            end
            aux_C_1 = exp(-rep_rho .* u_i); 
            aux_C_2 = exp(-rep_rho .* u_j);
            aux_C_3 = exp(-rep_rho);
            aux = 1 + (aux_C_1 - 1) .* (aux_C_2 - 1) ./ (aux_C_3 - 1); 
            C_aux = -log(aux) ./ rep_rho;
            D = D + signs(z_pos) * (u_i0 .* u_j0 .* C_aux);
          else 
            if i_zv == 0
              u_i0 = U_i';
            else
              u_i0 = M1_ones;
            end  
            if j_zv == 0
              u_j0 = U_j';
            else
              u_j0 = M1_ones;
            end
            D = D + signs(z_pos) * (u_i0 .* u_j0);
          end            
          
          % GRADIENT %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
          
          if false
              
          if G(y_i, y_j) == 1
            aux_C_bar = exp(-rep_rho .* (u_i + u_j));            
            aux_C_den = aux_C_3 + aux_C_bar - aux_C_1 - aux_C_2;
          end          
          if i_zv == 0 % Gradient for the loadings of Y_i
            if G(y_i, y_j) == 1
              C_bar = (aux_C_bar - aux_C_1) ./ aux_C_den;
              core_g = u_j0 .* ...
                         (((1 - exp_i) ./ u_i) .* C_aux + exp_i .* C_bar);
              rep_i_g1 = rep_u_i_g .* rep_X{p1};       
              rep_i_g2 = rep_u_i_g;
            else
              rep_i_g1 = (U_g_i .* X_v(:, p1))';
              rep_i_g2 = U_g_i';
              core_g = u_j0;
            end                                                              
            D_g{1}  = D_g{1} + signs(z_pos) * (rep_i_g1 .* core_g);          
            D_g{2}  = D_g{2} + signs(z_pos) * (rep_i_g2 .* core_g);
          end
          if j_zv == 0 % Gradient for the loadings of Y_j
            if G(y_i, y_j) == 1
              C_bar = (aux_C_bar - aux_C_2) ./ aux_C_den;
              core_g = u_i0 .* ...
                         (((1 - exp_j) ./ u_j) .* C_aux + exp_j .* C_bar);
              rep_j_g1 = rep_u_j_g .* rep_X{p2};       
              rep_j_g2 = rep_u_j_g;
            else
              rep_j_g1 = (U_g_j .* X_v(:, p2))';
              rep_j_g2 = U_g_j';
              core_g = u_i0;
            end                                                              
            D_g{3}  = D_g{3} + signs(z_pos) * (rep_j_g1 .* core_g);                      
            D_g{4}  = D_g{4} + signs(z_pos) * (rep_j_g2 .* core_g);
          end          
          
          end
          
          % ITERATE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
          
          z_pos = z_pos + 1;
          
        end % end of PMF transformation          

        pmf = mean(D * X_w_choice, 2);
        N_ij = Y_counts{v(1) + 1, v(2) + 1}(y_i, y_j);
        score = score + N_ij * (log(pmf' * w_iter)) * region_weight(p1, p2);
        
        if false
        pos1 = 2 * (y_i - 1); pos2 = 2 * (y_j - 1);
        for q = 1:2
          gr(pos1 + q) = gr(pos1 + q) + ...
                           N_ij * (((mean(D_g{q} * X_w_choice, 2)) ./ pmf)' * w_iter) * ...
                           region_weight(p1, p2);
          gr(pos2 + q) = gr(pos2 + q) + ...
                           N_ij * (((mean(D_g{2 + q} * X_w_choice, 2)) ./ pmf)' * w_iter) * ...
                           region_weight(p1, p2);
        end     
        end
        
      end % end of {Y_i, Y_j} calculations
      
    end
  end

  n_r = n_r + 1;
  
end
  
score = score / N;
gr = gr / N;

if isinf(score) || isnan(score)
  score = realmax - 1;
  gr = 0;
end
